@@ -2,7 +2,7 @@
 // detail/shared_ptr.hpp
 // ~~~~~~~~~~~~~~~~~~~~~
 //
-// Copyright (c) 2003-2011 Christopher M. Kohlhoff (chris at kohlhoff dot com)
+// Copyright (c) 2003-2016 Christopher M. Kohlhoff (chris at kohlhoff dot com)
 //
 // Distributed under the Boost Software License, Version 1.0. (See accompanying
 // file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
@@ -17,20 +17,20 @@
 
 #include "asio/detail/config.hpp"
 
-#if defined(_MSC_VER) && (_MSC_VER >= 1600)
+#if defined(ASIO_HAS_STD_SHARED_PTR)
 # include <memory>
-#else
+#else // defined(ASIO_HAS_STD_SHARED_PTR)
 # include <boost/shared_ptr.hpp>
-#endif
+#endif // defined(ASIO_HAS_STD_SHARED_PTR)
 
 namespace asio {
 namespace detail {
 
-#if defined(_MSC_VER) && (_MSC_VER >= 1600)
+#if defined(ASIO_HAS_STD_SHARED_PTR)
 using std::shared_ptr;
-#else
+#else // defined(ASIO_HAS_STD_SHARED_PTR)
 using boost::shared_ptr;
-#endif
+#endif // defined(ASIO_HAS_STD_SHARED_PTR)
 
 } // namespace detail
 } // namespace asio
