@@ -5,16 +5,20 @@
 #ifndef GALERA_CERTIFICATION_HPP
 #define GALERA_CERTIFICATION_HPP
 
+
+#include "nbo.hpp"
 #include "trx_handle.hpp"
 #include "key_entry_ng.hpp"
 #include "galera_service_thd.hpp"
+#include "galera_view.hpp"
 
-#include "gu_unordered.hpp"
-#include "gu_lock.hpp"
-#include "gu_config.hpp"
+#include <gu_shared_ptr.hpp>
+#include <gu_unordered.hpp>
+#include <gu_lock.hpp>
+#include <gu_config.hpp>
+#include <gu_gtid.hpp>
 
 #include <map>
-#include <set>
 #include <list>
 
 namespace galera
@@ -35,11 +39,15 @@ namespace galera
                                  KeyEntryPtrHashNG, KeyEntryPtrEqualNG>
         CertIndexNG;
 
+        typedef gu::UnorderedMultiset<KeyEntryNG*,
+                                      KeyEntryPtrHashNG, KeyEntryPtrEqualNG>
+        CertIndexNBO;
+
     private:
 
-        typedef std::multiset<wsrep_seqno_t>        DepsSet;
+        typedef std::multiset<wsrep_seqno_t>             DepsSet;
 
-        typedef std::map<wsrep_seqno_t, TrxHandle*> TrxMap;
+        typedef std::map<wsrep_seqno_t, TrxHandleSlavePtr> TrxMap;
 
     public:
 
@@ -49,13 +57,17 @@ namespace galera
             TEST_FAILED
         } TestResult;
 
-        Certification(gu::Config& conf, ServiceThd& thd);
+        Certification(gu::Config& conf, ServiceThd* thd);
         ~Certification();
 
-        void assign_initial_position(wsrep_seqno_t seqno, int versiono);
-        TestResult append_trx(TrxHandle*);
-        TestResult test(TrxHandle*, bool = true);
+        void assign_initial_position(const gu::GTID& gtid, int version);
+        TestResult append_trx(const TrxHandleSlavePtr&);
+        TestResult test(const TrxHandleSlavePtr&, bool store_keys);
         wsrep_seqno_t position() const { return position_; }
+        wsrep_seqno_t increment_position(); /* for dummy IST events */
+
+        /* this is for configuration change use */
+        void adjust_position(const View&, const gu::GTID& gtid, int version);
 
         wsrep_seqno_t
         get_safe_to_discard_seqno() const
@@ -78,8 +90,7 @@ namespace galera
 
         // Set trx corresponding to handle committed. Return purge seqno if
         // index purge is required, -1 otherwise.
-        wsrep_seqno_t set_trx_committed(TrxHandle*);
-        TrxHandle* get_trx(wsrep_seqno_t);
+        wsrep_seqno_t set_trx_committed(TrxHandleSlave&);
 
         // statistics section
         void stats_get(double& avg_cert_interval,
@@ -108,20 +119,44 @@ namespace galera
 
         void param_set(const std::string& key, const std::string& value);
 
+        wsrep_seqno_t lowest_trx_seqno() const
+        {
+            return (trx_map_.empty() ? position_ : trx_map_.begin()->first);
+        }
+
+        //
+        // NBO context lifecycle:
+        // * Context is created when NBO-start event is received
+        // * Context stays in nbo_ctx_map_ until client calls erase_nbo_ctx()
+        //
+
+        // Get NBO context matching to global seqno
+        gu::shared_ptr<NBOCtx>::type nbo_ctx(wsrep_seqno_t);
+        // Erase NBO context entry
+        void erase_nbo_ctx(wsrep_seqno_t);
+        size_t nbo_size() const { return nbo_map_.size(); }
+
+        void mark_inconsistent();
+        bool is_inconsistent() const { return inconsistent_; }
+
     private:
 
-        TestResult do_test(TrxHandle*, bool);
-        TestResult do_test_v1to2(TrxHandle*, bool);
-        TestResult do_test_v3to4(TrxHandle*, bool);
-        TestResult do_test_preordered(TrxHandle*);
-        void purge_for_trx(TrxHandle*);
-        void purge_for_trx_v1to2(TrxHandle*);
-        void purge_for_trx_v3(TrxHandle*);
+        // Non-copyable
+        Certification(const Certification&);
+        Certification& operator=(const Certification&);
+
+        TestResult do_test(const TrxHandleSlavePtr&, bool store_keys);
+        TestResult do_test_v3to5(TrxHandleSlave*, bool);
+        TestResult do_test_preordered(TrxHandleSlave*);
+        TestResult do_test_nbo(const TrxHandleSlavePtr&);
+        void purge_for_trx(TrxHandleSlave*);
 
         // unprotected variants for internal use
         wsrep_seqno_t get_safe_to_discard_seqno_() const;
         wsrep_seqno_t purge_trxs_upto_(wsrep_seqno_t, bool sync);
 
+        gu::shared_ptr<NBOCtx>::type nbo_ctx_unlocked(wsrep_seqno_t);
+
         bool index_purge_required()
         {
             static unsigned int const KEYS_THRESHOLD (1   << 10); // 1K
@@ -146,27 +181,37 @@ namespace galera
             void operator()(TrxMap::value_type& vt) const
             {
                 {
-                    TrxHandle* trx(vt.second);
-                    TrxHandleLock lock(*trx);
+                    TrxHandleSlave* trx(vt.second.get());
+                    // Trying to lock trx mutex here may cause deadlock
+                    // with streaming replication. Locking can be skipped
+                    // because trx is only read here and refcount uses atomics.
+                    // Memory barrier is provided by certification mutex.
+                    //
+                    // TrxHandleLock   lock(*trx);
 
-                    if (trx->is_committed() == false)
+                    if (!cert_.is_inconsistent())
                     {
-                        log_warn << "trx not committed in purge and discard: "
-                                 << *trx;
+                        assert(trx->is_committed() == true);
+                        if (trx->is_committed() == false)
+                        {
+                            log_warn <<"trx not committed in purge and discard: "
+                                     << *trx;
+                        }
                     }
 
-                    if (trx->depends_seqno() > -1)
+                    // If depends seqno is not WSREP_SEQNO_UNDEFINED
+                    // write set certification has passed and keys have been
+                    // inserted into index and purge is needed.
+                    // TOI write sets will always pass regular certification
+                    // and keys will be inserted, however if they fail
+                    // NBO certification depends seqno is set to
+                    // WSREP_SEQNO_UNDEFINED. Therefore purge should always
+                    // be done for TOI write sets.
+                    if (trx->depends_seqno() >= 0 || trx->is_toi() == true)
                     {
                         cert_.purge_for_trx(trx);
                     }
-
-                    if (trx->refcnt() > 1)
-                    {
-                        log_debug << "trx "     << trx->trx_id()
-                                  << " refcnt " << trx->refcnt();
-                    }
                 }
-                vt.second->unref();
             }
 
             PurgeAndDiscard(const PurgeAndDiscard& other) : cert_(other.cert_)
@@ -181,14 +226,19 @@ namespace galera
         int           version_;
         gu::Config&   conf_;
         TrxMap        trx_map_;
-        CertIndex     cert_index_;
         CertIndexNG   cert_index_ng_;
+        NBOMap        nbo_map_;
+        NBOCtxMap     nbo_ctx_map_;
+        CertIndexNBO  nbo_index_;
+        TrxHandleSlave::Pool nbo_pool_;
         DepsSet       deps_set_;
-        ServiceThd&   service_thd_;
+        View          current_view_;
+        ServiceThd*   service_thd_;
         gu::Mutex     mutex_;
         size_t        trx_size_warn_count_;
         wsrep_seqno_t initial_position_;
         wsrep_seqno_t position_;
+        wsrep_seqno_t nbo_position_;
         wsrep_seqno_t safe_to_discard_seqno_;
         wsrep_seqno_t last_pa_unsafe_;
         wsrep_seqno_t last_preordered_seqno_;
@@ -212,6 +262,7 @@ namespace galera
 
         unsigned int const max_length_check_; /* Mask how often to check */
 
+        bool               inconsistent_;
         bool               log_conflicts_;
         bool               optimistic_pa_;
     };
