@@ -1,17 +1,18 @@
 /*
- * Copyright (C) 2012-2020 Codership Oy <info@codership.com>
+ * Copyright (C) 2012-2017 Codership Oy <info@codership.com>
  */
 
 #include "../src/saved_state.hpp"
 
-#include "../src/uuid.hpp"
-
-#include "gu_inttypes.hpp"
+#include <gu_uuid.hpp>
 
 #include <check.h>
 #include <errno.h>
 #include <gu_threads.h>
 
+#define __STDC_FORMAT_MACROS
+#include <inttypes.h>
+
 static volatile bool stop(false);
 
 using namespace galera;
@@ -33,11 +34,6 @@ thread_routine (void* arg)
 
 static const int max_threads(16);
 static gu_thread_t threads[max_threads];
-#if defined(GALERA_WITH_VALGRIND)
-static const int iterations(10);
-#else
-static const int iterations(100);
-#endif // GALERA_WITH_VALGRIND
 
 static void
 start_threads(void* arg)
@@ -48,8 +44,8 @@ start_threads(void* arg)
     {
         gu_thread_t t;
         int err = gu_thread_create (&t, NULL, thread_routine, arg);
-        ck_assert_msg(0 == err, "Failed to start thread %d: %d (%s)",
-                      ret, err, strerror(err));
+        fail_if (err, "Failed to start thread %d: %d (%s)",
+                 ret, err, strerror(err));
         threads[ret] = t;
     }
 }
@@ -71,8 +67,7 @@ START_TEST(test_basic)
 {
     unlink (fname);
 
-    union { wsrep_uuid_t uuid; gu_word_t align; } aligned;
-    wsrep_uuid_t& uuid(aligned.uuid);
+    wsrep_uuid_t  uuid;
     wsrep_seqno_t seqno;
     bool safe_to_bootstrap;
 
@@ -81,12 +76,11 @@ START_TEST(test_basic)
 
         st.get(uuid, seqno, safe_to_bootstrap);
 
-        ck_assert(uuid  == WSREP_UUID_UNDEFINED);
-        ck_assert(seqno == WSREP_SEQNO_UNDEFINED);
-        ck_assert(safe_to_bootstrap == true);
+        fail_if (uuid  != WSREP_UUID_UNDEFINED);
+        fail_if (seqno != WSREP_SEQNO_UNDEFINED);
+        fail_if (safe_to_bootstrap != true);
 
-        gu_uuid_from_string("b2c01654-8dfe-11e1-0800-a834d641cfb5",
-                            to_gu_uuid(uuid));
+        gu_uuid_from_string("b2c01654-8dfe-11e1-0800-a834d641cfb5", uuid);
         seqno = 2345234LL;
 
         st.set(uuid, seqno, false);
@@ -101,9 +95,9 @@ START_TEST(test_basic)
 
         st.get(u, s, stb);
 
-        ck_assert(u == uuid);
-        ck_assert(s == seqno);
-        ck_assert(stb == false);
+        fail_if (u != uuid);
+        fail_if (s != seqno);
+        fail_if (stb != false);
     }
 }
 END_TEST
@@ -114,20 +108,19 @@ START_TEST(test_unsafe)
 {
     SavedState st(fname);
 
-    union { wsrep_uuid_t uuid; gu_word_t align; } aligned;
-    wsrep_uuid_t& uuid(aligned.uuid);
+    wsrep_uuid_t  uuid;
     wsrep_seqno_t seqno;
     bool safe_to_bootstrap;
 
     st.get(uuid, seqno, safe_to_bootstrap);
 
-    ck_assert(uuid  != WSREP_UUID_UNDEFINED);
-    ck_assert(seqno != WSREP_SEQNO_UNDEFINED);
-    ck_assert(safe_to_bootstrap != true);
+    fail_if (uuid  == WSREP_UUID_UNDEFINED);
+    fail_if (seqno == WSREP_SEQNO_UNDEFINED);
+    fail_if (safe_to_bootstrap == true);
 
     st.set(uuid, WSREP_SEQNO_UNDEFINED, false);
 
-    for (int i = 0; i < iterations; ++i)
+    for (int i = 0; i < 100; ++i)
     {
         start_threads(&st);
         mark_point();
@@ -139,9 +132,9 @@ START_TEST(test_unsafe)
         mark_point();
         st.get(uuid, seqno, safe_to_bootstrap);
 
-        ck_assert(uuid != WSREP_UUID_UNDEFINED);
-        ck_assert(seqno == i);
-        ck_assert(safe_to_bootstrap == false);
+        fail_if (uuid == WSREP_UUID_UNDEFINED);
+        fail_if (seqno != i);
+        fail_if (safe_to_bootstrap != false);
     }
 
     long marks, locks, writes;
@@ -157,8 +150,7 @@ END_TEST
 
 START_TEST(test_corrupt)
 {
-    union { wsrep_uuid_t uuid; gu_word_t align; } aligned;
-    wsrep_uuid_t& uuid(aligned.uuid);
+    wsrep_uuid_t  uuid;
     wsrep_seqno_t seqno;
     bool safe_to_bootstrap;
 
@@ -167,16 +159,16 @@ START_TEST(test_corrupt)
 
         st.get(uuid, seqno, safe_to_bootstrap);
 
-        ck_assert(uuid  != WSREP_UUID_UNDEFINED);
-        ck_assert(seqno != WSREP_SEQNO_UNDEFINED);
-        ck_assert(safe_to_bootstrap != true);
+        fail_if (uuid  == WSREP_UUID_UNDEFINED);
+        fail_if (seqno == WSREP_SEQNO_UNDEFINED);
+        fail_if (safe_to_bootstrap == true);
 
         st.set(uuid, WSREP_SEQNO_UNDEFINED, false);
     }
 
     long marks(0), locks(0), writes(0);
 
-    for (int i = 0; i < iterations; ++i)
+    for (int i = 0; i < 100; ++i)
     {
         SavedState st(fname);
         // explicitly overwrite corruption mark.
@@ -198,9 +190,9 @@ START_TEST(test_corrupt)
         st.get(u, s, stb);
 
         // make sure that mark_corrupt() stays
-        ck_assert(u == WSREP_UUID_UNDEFINED);
-        ck_assert(s == WSREP_SEQNO_UNDEFINED);
-        ck_assert(stb == false);
+        fail_if (u != WSREP_UUID_UNDEFINED);
+        fail_if (s != WSREP_SEQNO_UNDEFINED);
+        fail_if (stb != false);
 
         long m, l, w;
 
@@ -220,6 +212,9 @@ START_TEST(test_corrupt)
 }
 END_TEST
 
+#define WAIT_FOR(cond)                                                  \
+    { int count = 1000; while (--count && !(cond)) { usleep (TEST_USLEEP); }}
+
 Suite* saved_state_suite()
 {
     Suite* s = suite_create ("saved_state");
