@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2020 Codership Oy <info@codership.com>
+ * Copyright (C) 2009-2018 Codership Oy <info@codership.com>
  */
 
 #include "GCache.hpp"
@@ -22,13 +22,12 @@ namespace gcache
         mallocs  = 0;
         reallocs = 0;
 
-        gid            = gu::UUID();
+        seqno_locked   = SEQNO_NONE;
         seqno_max      = SEQNO_NONE;
         seqno_released = SEQNO_NONE;
-        seqno_locked   = SEQNO_MAX;
-        seqno_locked_count = 0;
+        gid            = gu::UUID();
 
-        seqno2ptr.clear(SEQNO_NONE);
+        seqno2ptr.clear();
 
 #ifndef NDEBUG
         buf_tracker.clear();
@@ -40,7 +39,8 @@ namespace gcache
         config    (cfg),
         params    (config, data_dir),
         mtx       (),
-        seqno2ptr (SEQNO_NONE),
+        cond      (),
+        seqno2ptr (),
         gid       (),
         mem       (params.mem_size(), seqno2ptr, params.debug()),
         rb        (params.rb_name(), params.rb_size(), seqno2ptr, gid,
@@ -54,11 +54,10 @@ namespace gcache
         mallocs   (0),
         reallocs  (0),
         frees     (0),
-        seqno_max     (seqno2ptr.empty() ?
-                       SEQNO_NONE : seqno2ptr.index_back()),
-        seqno_released(seqno_max),
-        seqno_locked  (SEQNO_MAX),
-        seqno_locked_count(0)
+        seqno_locked(SEQNO_NONE),
+        seqno_max   (seqno2ptr.empty() ?
+                     SEQNO_NONE : seqno2ptr.rbegin()->first),
+        seqno_released(seqno_max)
 #ifndef NDEBUG
         ,buf_tracker()
 #endif
@@ -114,24 +113,3 @@ int64_t gcache_seqno_min (gcache_t* gc)
     gcache::GCache* gcache = reinterpret_cast<gcache::GCache*>(gc);
     return gcache->seqno_min ();
 }
-
-#if DEPRECATED
-void  gcache_seqno_init   (gcache_t* gc, int64_t seqno)
-{
-    gcache::GCache* gcache = reinterpret_cast<gcache::GCache*>(gc);
-    gcache->seqno_init (seqno);
-}
-
-void  gcache_seqno_assign(gcache_t* gc, const void* ptr, int64_t seqno)
-{
-    gcache::GCache* gcache = reinterpret_cast<gcache::GCache*>(gc);
-    gcache->seqno_assign (ptr, seqno, -1, false);
-}
-
-void  gcache_seqno_release(gcache_t* gc, const void* ptr)
-{
-    gcache::GCache* gcache = reinterpret_cast<gcache::GCache*>(gc);
-    gcache->seqno_release ();
-}
-#endif
-
