@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2020 Codership Oy <info@codership.com>
+ * Copyright (C) 2009-2018 Codership Oy <info@codership.com>
  */
 
 #ifndef __GCACHE_H__
@@ -13,6 +13,7 @@
 #include <gu_types.hpp>
 #include <gu_lock.hpp> // for gu::Mutex and gu::Cond
 #include <gu_config.hpp>
+#include <gu_gtid.hpp>
 
 #include <string>
 #include <iostream>
@@ -27,6 +28,8 @@ namespace gcache
     {
     public:
 
+        static const std::string& PARAMS_DIR;
+
         static void register_params(gu::Config& cfg)
         {
             Params::register_params(cfg);
@@ -58,74 +61,84 @@ namespace gcache
          * Reinitialize seqno sequence (after SST or such)
          * Clears seqno->ptr map // and sets seqno_min to seqno.
          */
-        void  seqno_reset (const gu::UUID& gid, seqno_t seqno);
+        void seqno_reset (const gu::GTID& gtid);
 
         /*!
          * Assign sequence number to buffer pointed to by ptr
          */
-        void  seqno_assign (const void* ptr,
-                            seqno_t     seqno_g,
-                            seqno_t     seqno_d);
+        void seqno_assign (const void* ptr,
+                           int64_t     seqno_g,
+                           uint8_t     type,
+                           bool        skip);
+
+        /*!
+         * Mark buffer to be skipped
+         */
+        void seqno_skip (const void* ptr,
+                         int64_t     seqno_g,
+                         uint8_t     type);
 
         /*!
          * Release (free) buffers up to seqno
          */
-        void seqno_release (seqno_t seqno);
+        void seqno_release (int64_t seqno);
 
         /*!
          * Returns smallest seqno present in history
          */
-        seqno_t seqno_min() const
+        int64_t seqno_min() const
         {
             gu::Lock lock(mtx);
             if (gu_likely(!seqno2ptr.empty()))
-                return seqno2ptr.index_begin();
+                return seqno2ptr.begin()->first;
             else
-                return SEQNO_ILL;
+                return -1;
         }
 
         /*!
          * Move lock to a given seqno.
          * @throws gu::NotFound if seqno is not in the cache.
          */
-        void  seqno_lock (seqno_t const seqno_g);
+        void  seqno_lock (int64_t const seqno_g);
 
-        /*!          DEPRECATED
+        /*!
          * Get pointer to buffer identified by seqno.
-         * Moves lock to the given seqno.
+         * Moves lock to the given seqno and clears released flag if any.
+         * The buffer will need to be "freed" again.
          * @throws NotFound
          */
-        const void* seqno_get_ptr (seqno_t  seqno_g,
-                                   seqno_t& seqno_d,
-                                   ssize_t& size);
+        const void* seqno_get_ptr (int64_t seqno_g, ssize_t& size);
 
         class Buffer
         {
         public:
 
-            Buffer() : seqno_g_(), seqno_d_(), ptr_(), size_() { }
+            Buffer() : seqno_g_(), ptr_(), size_(), skip_(), type_() { }
 
             Buffer (const Buffer& other)
                 :
                 seqno_g_(other.seqno_g_),
-                seqno_d_(other.seqno_d_),
                 ptr_    (other.ptr_),
-                size_   (other.size_)
+                size_   (other.size_),
+                skip_   (other.skip_),
+                type_   (other.type_)
             { }
 
             Buffer& operator= (const Buffer& other)
             {
                 seqno_g_ = other.seqno_g_;
-                seqno_d_ = other.seqno_d_;
                 ptr_     = other.ptr_;
                 size_    = other.size_;
+                skip_    = other.skip_;
+                type_    = other.type_;
                 return *this;
             }
 
-            seqno_t           seqno_g() const { return seqno_g_; }
-            seqno_t           seqno_d() const { return seqno_d_; }
+            int64_t           seqno_g() const { return seqno_g_; }
             const gu::byte_t* ptr()     const { return ptr_;     }
             ssize_type        size()    const { return size_;    }
+            bool              skip()    const { return skip_;    }
+            uint8_t           type()    const { return type_;    }
 
         protected:
 
@@ -134,18 +147,19 @@ namespace gcache
                 ptr_ = reinterpret_cast<const gu::byte_t*>(p);
             }
 
-            void set_other (int64_t g, int64_t d, ssize_type s)
+            void set_other (int64_t g, ssize_type s, bool skp, uint8_t t)
             {
                 assert(s > 0);
-                seqno_g_ = g; seqno_d_ = d; size_ = s;
+                seqno_g_ = g; size_ = s; skip_ = skp, type_ = t;
             }
 
         private:
 
-            seqno_t           seqno_g_;
-            seqno_t           seqno_d_;
+            int64_t           seqno_g_;
             const gu::byte_t* ptr_;
-            ssize_type        size_; /* same type as passed to malloc() */
+            ssize_type        size_;
+            bool              skip_;
+            uint8_t           type_;
 
             friend class GCache;
         };
@@ -157,7 +171,7 @@ namespace gcache
          *
          * @retval number of buffers filled (<= v.size())
          */
-        size_t seqno_get_buffers (std::vector<Buffer>& v, seqno_t start);
+        size_t seqno_get_buffers (std::vector<Buffer>& v, int64_t start);
 
         /*!
          * Releases any seqno locks present.
@@ -216,6 +230,7 @@ namespace gcache
             params;
 
         gu::Mutex       mtx;
+        gu::Cond        cond;
 
         seqno2ptr_t     seqno2ptr;
         gu::UUID        gid;
@@ -228,11 +243,9 @@ namespace gcache
         long long       reallocs;
         long long       frees;
 
-        seqno_t         seqno_max;
-        seqno_t         seqno_released;
-
-        seqno_t         seqno_locked;
-        int             seqno_locked_count;
+        int64_t         seqno_locked;
+        int64_t         seqno_max;
+        int64_t         seqno_released;
 
 #ifndef NDEBUG
         std::set<const void*> buf_tracker;
@@ -241,10 +254,10 @@ namespace gcache
         void discard_buffer (BufferHeader* bh);
 
         /* returns true when successfully discards all seqnos up to s */
-        bool discard_seqno (seqno_t s);
+        bool discard_seqno (int64_t s);
 
         /* discards all seqnos greater than s */
-        void discard_tail (seqno_t s);
+        void discard_tail (int64_t s);
 
         // disable copying
         GCache (const GCache&);
