@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2021 Codership Oy <info@codership.com>
+ * Copyright (C) 2009-2019 Codership Oy <info@codership.com>
  */
 
 #include "GCache.hpp"
@@ -24,40 +24,32 @@ namespace gcache
     }
 
     bool
-    GCache::discard_seqno (seqno_t seqno)
+    GCache::discard_seqno (int64_t seqno)
     {
+        assert(mtx.locked() && mtx.owned());
+
 #ifndef NDEBUG
-        seqno_t const begin(params.debug() ?
-                            (seqno2ptr.empty() ?
-                             SEQNO_NONE : seqno2ptr.index_begin()) : SEQNO_NONE);
+        seqno_t begin(0);
         if (params.debug())
         {
+            begin = (seqno2ptr.begin() != seqno2ptr.end() ?
+                     seqno2ptr.begin()->first : 0);
             assert(begin > 0);
             log_info << "GCache::discard_seqno(" << begin << " - "
                      << seqno << ")";
         }
 #endif
-        /* if we can't complete the operation, let's not even start */
-        if (seqno >= seqno_locked)
+        for (seqno2ptr_t::iterator i = seqno2ptr.begin();
+             i != seqno2ptr.end() && i->first <= seqno;)
         {
-#ifndef NDEBUG
-            if (params.debug())
-            {
-                log_info << "GCache::discard_seqno(" << begin << " - " << seqno
-                         << "): " << seqno_locked << " is locked, bailing out.";
-            }
-#endif
-            return false;
-        }
-
-        while (seqno2ptr.index_begin() <= seqno && !seqno2ptr.empty())
-        {
-            BufferHeader* const bh(ptr2BH(seqno2ptr.front()));
+            BufferHeader* bh(ptr2BH (i->second));
 
             if (gu_likely(BH_is_released(bh)))
             {
-                assert (bh->seqno_g == seqno2ptr.index_begin());
+                assert (bh->seqno_g == i->first);
                 assert (bh->seqno_g <= seqno);
+
+                seqno2ptr.erase (i++); // post ++ is significant!
                 discard_buffer(bh);
             }
             else
@@ -72,25 +64,26 @@ namespace gcache
 #endif
                 return false;
             }
-
-            seqno2ptr.pop_front();
         }
 
         return true;
     }
 
     void
-    GCache::discard_tail (seqno_t const seqno)
+    GCache::discard_tail (int64_t seqno)
     {
-        while (seqno2ptr.index_back() > seqno && !seqno2ptr.empty())
+        seqno2ptr_t::reverse_iterator r;
+        while ((r = seqno2ptr.rbegin()) != seqno2ptr.rend() &&
+               r->first > seqno)
         {
-            BufferHeader* bh(ptr2BH(seqno2ptr.back()));
+            BufferHeader* bh(ptr2BH(r->second));
 
             assert(BH_is_released(bh));
-            assert(bh->seqno_g == seqno2ptr.index_back());
+            assert(bh->seqno_g == r->first);
+            assert(bh->seqno_g > seqno);
 
+            seqno2ptr.erase(--(seqno2ptr.end()));
             discard_buffer(bh);
-            seqno2ptr.pop_back();
         }
     }
 
@@ -141,9 +134,8 @@ namespace gcache
             {
                 log_fatal << "OOO release: seqno_released " << seqno_released
                           << ", releasing " << bh->seqno_g;
+                assert(0);
             }
-            assert(seqno_released + 1 == bh->seqno_g ||
-                   SEQNO_NONE == seqno_released);
 #endif
             new_released = bh->seqno_g;
         }
@@ -168,7 +160,7 @@ namespace gcache
             {
                 if (gu_unlikely(!discard_seqno(bh->seqno_g)))
                 {
-                    new_released = (bh->seqno_g - 1);
+                    new_released = (seqno2ptr.begin()->first - 1);
                     assert(seqno_released <= new_released);
                 }
             }
@@ -182,6 +174,13 @@ namespace gcache
         }
         rb.assert_size_free();
 
+#ifndef NDEBUG
+        if (params.debug())
+        {
+            log_info << "GCache::free_common(): seqno_released: "
+                     << seqno_released << " -> " << new_released;
+        }
+#endif
         seqno_released = new_released;
     }
 
@@ -195,16 +194,8 @@ namespace gcache
 
 #ifndef NDEBUG
             if (params.debug()) { log_info << "GCache::free() " << bh; }
-            seqno_t const old_sr(seqno_released);
 #endif
             free_common (bh);
-#ifndef NDEBUG
-            if (params.debug())
-            {
-                log_info << "GCache::free() seqno_released: "
-                         << old_sr << " -> " << seqno_released;
-            }
-#endif
         }
         else {
             log_warn << "Attempt to free a null pointer";
