@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2020 Codership Oy <info@codership.com>
+ * Copyright (C) 2010-2018 Codership Oy <info@codership.com>
  */
 
 /*! @file ring buffer storage class */
@@ -34,16 +34,31 @@ namespace gcache
 
         void* malloc  (size_type size);
 
+        void* realloc (void* ptr, size_type size);
+
         void  free    (BufferHeader* bh);
 
-        void* realloc (void* ptr, size_type size);
+        void repossess(BufferHeader* bh)
+        {
+            assert(bh->size > 0);
+            assert(bh->seqno_g != SEQNO_NONE);
+            assert(bh->store == BUFFER_IN_RB);
+            assert(bh->ctx == reinterpret_cast<BH_ctx_t>(this));
+            assert(BH_is_released(bh)); // will be marked unreleased by caller
+
+            size_used_ += bh->size;
+            assert(size_used_ <= size_cache_);
+        }
 
         void  discard (BufferHeader* const bh)
         {
             assert (BH_is_released(bh));
-            assert (SEQNO_ILL == bh->seqno_g);
+            assert (BUFFER_IN_RB == bh->store);
+
             size_free_ += bh->size;
             assert (size_free_ <= size_cache_);
+
+            bh->seqno_g = SEQNO_ILL;
         }
 
         size_t size      () const { return size_cache_; }
@@ -114,15 +129,6 @@ namespace gcache
 
         void set_debug(int const dbg) { debug_ = dbg & DEBUG; }
 
-#ifdef GCACHE_RB_UNIT_TEST
-        ptrdiff_t offset(const void* const ptr) const
-        {
-            return static_cast<const uint8_t*>(ptr) - start_;
-        }
-#endif
-
-        void dump_map() const;
-
     private:
 
         static size_t const PREAMBLE_LEN = 1024;
@@ -173,7 +179,7 @@ namespace gcache
         void          close_preamble();
 
         // returns lower bound (not inclusive) of valid seqno range
-        seqno_t       scan(off_t offset, int scan_step);
+        int64_t       scan(off_t offset, int scan_step);
         void          recover(off_t offset, int version);
 
         void          estimate_space();
