@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2011-2020 Codership Oy <info@codership.com>
+ * Copyright (C) 2011-2018 Codership Oy <info@codership.com>
  *
  * $Id$
  */
@@ -32,46 +32,45 @@ START_TEST(test1)
 
     size_t const rb_size(ALLOC_SIZE(2) * 2);
 
-    seqno2ptr_t s2p(SEQNO_NONE);
+    std::map<int64_t, const void*> s2p;
     gu::UUID   gid(GID);
     RingBuffer rb(RB_NAME, rb_size, s2p, gid, 0, false);
 
-    ck_assert_msg(rb.size() == rb_size,
-                  "Expected %zd, got %zd", rb_size, rb.size());
+    fail_if (rb.size() != rb_size, "Expected %zd, got %zd", rb_size, rb.size());
 
     if (gid != GID)
     {
         std::ostringstream os;
         os << "Expected GID: " << GID << ", got: " << gid;
-        ck_abort_msg("%s", os.str().c_str());
+        fail(os.str().c_str());
     }
 
     void* buf1 = rb.malloc (MemOps::align_size(rb_size/2 + 1));
-    ck_assert(NULL == buf1); // > 1/2 size
+    fail_if (NULL != buf1); // > 1/2 size
 
     buf1 = rb.malloc (ALLOC_SIZE(1));
-    ck_assert(NULL != buf1);
+    fail_if (NULL == buf1);
 
     BufferHeader* bh1(ptr2BH(buf1));
-    ck_assert(bh1->seqno_g == SEQNO_NONE);
-    ck_assert(!BH_is_released(bh1));
+    fail_if (bh1->seqno_g != SEQNO_NONE);
+    fail_if (BH_is_released(bh1));
 
     void* buf2 = rb.malloc (ALLOC_SIZE(2));
-    ck_assert(NULL != buf2);
-    ck_assert(!BH_is_released(bh1));
+    fail_if (NULL == buf2);
+    fail_if (BH_is_released(bh1));
 
     BufferHeader* bh2(ptr2BH(buf2));
-    ck_assert(bh2->seqno_g == SEQNO_NONE);
-    ck_assert(!BH_is_released(bh2));
+    fail_if (bh2->seqno_g != SEQNO_NONE);
+    fail_if (BH_is_released(bh2));
 
     void* tmp = rb.realloc (buf1, ALLOC_SIZE(2));
     // anything <= MemOps::ALIGNMENT should fit into original buffer
-    ck_assert(tmp == buf1 && MemOps::ALIGNMENT > 1);
+    fail_if(tmp != buf1 && MemOps::ALIGNMENT > 1);
 
     tmp = rb.realloc (buf1, ALLOC_SIZE(MemOps::ALIGNMENT + 1));
     // should require new buffer for which there's no space
-    ck_assert(bh2->seqno_g == SEQNO_NONE);
-    ck_assert(NULL == tmp);
+    fail_if (bh2->seqno_g != SEQNO_NONE);
+    fail_if (NULL != tmp);
 
     BH_release(bh2);
     rb.free (bh2);
@@ -79,34 +78,34 @@ START_TEST(test1)
     tmp = rb.realloc (buf1, ALLOC_SIZE(3));
     if (MemOps::ALIGNMENT > 2)
     {
-        ck_assert(NULL != tmp);
-        ck_assert(buf1 == tmp);
+        fail_if (NULL == tmp);
+        fail_if (buf1 != tmp);
     }
     else
     {
-        ck_assert(NULL == tmp);
+        fail_if (NULL != tmp);
     }
 
     BH_release(bh1);
     rb.free (bh1);
-    ck_assert(BH_is_released(bh1));
+    fail_if (!BH_is_released(bh1));
 
     buf1 = rb.malloc(ALLOC_SIZE(1));
-    ck_assert(NULL != buf1);
+    fail_if (NULL == buf1);
 
     tmp = rb.realloc (buf1, ALLOC_SIZE(2));
-    ck_assert(NULL != tmp);
-    ck_assert(tmp  == buf1);
+    fail_if (NULL == tmp);
+    fail_if (tmp != buf1);
 
     buf2 = rb.malloc (ALLOC_SIZE(1));
-    ck_assert(NULL != buf2);
+    fail_if (NULL == buf2);
 
     tmp = rb.realloc (buf2, ALLOC_SIZE(2));
-    ck_assert(NULL != tmp);
-    ck_assert(tmp  == buf2);
+    fail_if (NULL == tmp);
+    fail_if (tmp != buf2);
 
     tmp = rb.malloc (ALLOC_SIZE(1));
-    ck_assert(NULL == tmp);
+    fail_if (NULL != tmp);
 
     BH_release(ptr2BH(buf1));
     rb.free(ptr2BH(buf1));
@@ -115,7 +114,7 @@ START_TEST(test1)
     rb.free(ptr2BH(buf2));
 
     tmp = rb.malloc (ALLOC_SIZE(2));
-    ck_assert(NULL != tmp);
+    fail_if (NULL == tmp);
 
     mark_point();
 }
@@ -157,18 +156,24 @@ START_TEST(recovery)
         RingBuffer   rb;
 
         rb_ctx(size_t s, bool recover = true) :
-            size(s), s2p(SEQNO_NONE), gid(GID),
-            rb(RB_NAME, size, s2p, gid, 0, recover)
-        {}
+            size(s), s2p(), gid(GID), rb(RB_NAME, size, s2p, gid, 0, recover) {}
 
         void seqno_assign (seqno2ptr_t& s2p, void* const ptr,
                            seqno_t const g, seqno_t const d)
         {
-            s2p.insert(g, ptr);
+            const std::pair<seqno2ptr_iter_t, bool>& res
+                (s2p.insert(seqno2ptr_pair_t(g, ptr)));
+
+            if (false == res.second)
+            {
+                gu_throw_fatal <<"Attempt to reuse the same seqno: " << g
+                               <<". New ptr = " << ptr << ", previous ptr = "
+                               << res.first->second;
+            }
 
             BufferHeader* bh(ptr2BH(ptr));
             bh->seqno_g = g;
-            bh->seqno_d = d;
+            if (d < 0) bh->flags |= BUFFER_SKIPPED;
         }
 
         void* add_msg(struct msg& m)
@@ -194,8 +199,8 @@ START_TEST(recovery)
             os << "S2P map:\n";
             for (seqno2ptr_t::iterator i = s2p.begin(); i != s2p.end(); ++i)
             {
-                log_info << "\tseqno: " << s2p.index(i) << ", msg: "
-                         << reinterpret_cast<const char*>(*i) << "\n";
+                log_info << "\tseqno: " << i->first << ", msg: "
+                         << reinterpret_cast<const char*>(i->second) << "\n";
             }
 
             log_info << os.str();
@@ -208,37 +213,37 @@ START_TEST(recovery)
     {
         rb_ctx ctx(rb_5size);
 
-        ck_assert_msg(ctx.rb.size() == ctx.size,
-                      "Expected %zd, got %zd", ctx.size, ctx.rb.size());
+        fail_if (ctx.rb.size() != ctx.size,
+                 "Expected %zd, got %zd", ctx.size, ctx.rb.size());
 
         if (ctx.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(ctx.s2p.empty());
+        fail_if(!ctx.s2p.empty());
 
         void* m(ctx.add_msg(msgs[0]));
-        ck_assert(NULL != m);
-        ck_assert(*ctx.s2p.find(msgs[0].g) == m);
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[0].g)->second != m);
 
         m = ctx.add_msg(msgs[1]);
-        ck_assert(NULL != m);
-        ck_assert(*ctx.s2p.find(msgs[1].g) == m);
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[1].g)->second != m);
 
         m = ctx.add_msg(msgs[2]);
-        ck_assert(NULL != m);
-        ck_assert(*ctx.s2p.find(msgs[2].g) == m);
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[2].g)->second != m);
 
         m = ctx.add_msg(msgs[3]);
-        ck_assert(NULL != m);
-        ck_assert(msgs[3].g <= 0);
-        ck_assert(ctx.s2p.find(msgs[3].g) == ctx.s2p.end());
+        fail_if (NULL == m);
+        fail_if (msgs[3].g > 0);
+        fail_if (ctx.s2p.find(msgs[3].g) != ctx.s2p.end());
 
-        seqno_min = ctx.s2p.index_front();
-        seqno_max = ctx.s2p.index_back();
+        seqno_min = ctx.s2p.begin()->first;
+        seqno_max = ctx.s2p.rbegin()->first;
     }
 
     /* What we have now is |111222444***|----| */
@@ -249,84 +254,83 @@ START_TEST(recovery)
     {
         rb_ctx ctx(rb_5size);
 
-        ck_assert_msg(ctx.rb.size() == ctx.size,
-                      "Expected %zd, got %zd", ctx.size, ctx.rb.size());
+        fail_if (ctx.rb.size() != ctx.size,
+                 "Expected %zd, got %zd", ctx.size, ctx.rb.size());
 
         if (ctx.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(!ctx.s2p.empty());
-        ck_assert(ctx.s2p.size() == 1);
-        ck_assert(ctx.s2p.index_front() != seqno_min);
-        ck_assert(ctx.s2p.index_front() == seqno_max);
+        fail_if(ctx.s2p.empty());
+        fail_if(ctx.s2p.size() != 1);
+        fail_if(ctx.s2p.begin()->first == seqno_min);
+        fail_if(ctx.s2p.begin()->first != seqno_max);
 
         void* m(ctx.add_msg(msgs[4]));
-        ck_assert(NULL != m);
-        ck_assert(*ctx.s2p.find(msgs[4].g) == m);
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[4].g)->second != m);
 
         m = ctx.add_msg(msgs[5]);
-        ck_assert(NULL != m);
-        ck_assert(msgs[5].g <= 0);
-        ck_assert(ctx.s2p.find(msgs[5].g) == ctx.s2p.end());
+        fail_if (NULL == m);
+        fail_if (msgs[5].g > 0);
+        fail_if (ctx.s2p.find(msgs[5].g) != ctx.s2p.end());
 
         m = ctx.add_msg(msgs[6]);
-        ck_assert(NULL != m);
-        ck_assert(*ctx.s2p.find(msgs[6].g) <= m);
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[6].g)->second != m);
         // here we should have rollover
-        ck_assert(ptr2BH(m) == BH_cast(ctx.rb.start()));
+        fail_if (ptr2BH(m) != BH_cast(ctx.rb.start()));
 
-        seqno_min = ctx.s2p.index_front();
-        seqno_max = ctx.s2p.index_back();
+        seqno_min = ctx.s2p.begin()->first;
+        seqno_max = ctx.s2p.rbegin()->first;
     }
 
     /* What we have now is |555|---|444333***| */
     /* Reopening of the file should:
-     * 1) discard unordered message at the end
+     * 1) discard discard unordered message at the end
      * 2) continuous seqno interval should be now 3,4,5
      */
     {
         rb_ctx ctx0(rb_5size);
 
-        ck_assert_msg(ctx0.rb.size() == ctx0.size,
-                      "Expected %zd, got %zd", ctx0.size, ctx0.rb.size());
+        fail_if (ctx0.rb.size() != ctx0.size,
+                 "Expected %zd, got %zd", ctx0.size, ctx0.rb.size());
 
         if (ctx0.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx0.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(!ctx0.s2p.empty());
-        ck_assert(ctx0.s2p.size() == 3);
-        ck_assert(ctx0.s2p.index_front() == seqno_min);
-        ck_assert(ctx0.s2p.index_back()  == seqno_max);
+        fail_if(ctx0.s2p.empty());
+        fail_if(ctx0.s2p.size() != 3);
+        fail_if(ctx0.s2p.begin()->first  != seqno_min);
+        fail_if(ctx0.s2p.rbegin()->first != seqno_max);
 
         /* now try to open unclosed file. Results should be the same */
         rb_ctx ctx(rb_5size);
 
-        ck_assert_msg(ctx.rb.size() == ctx.size,
-                      "Expected %zd, got %zd", ctx.size, ctx.rb.size());
+        fail_if (ctx.rb.size() != ctx.size,
+                 "Expected %zd, got %zd", ctx.size, ctx.rb.size());
 
         if (ctx.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(!ctx.s2p.empty());
-        ck_assert_msg(ctx.s2p.size() == 3,
-                      "Expected seqno2ptr size 3, got %zd", ctx.s2p.size());
-        ck_assert(ctx.s2p.index_front() == seqno_min);
-        ck_assert(ctx.s2p.index_back()  == seqno_max);
+        fail_if(ctx.s2p.empty());
+        fail_if(ctx.s2p.size() != 3);
+        fail_if(ctx.s2p.begin()->first  != seqno_min);
+        fail_if(ctx.s2p.rbegin()->first != seqno_max);
 
-        seqno_min = ctx.s2p.index_front();
-        seqno_max = ctx.s2p.index_back();
+        seqno_min = ctx.s2p.begin()->first;
+        seqno_max = ctx.s2p.rbegin()->first;
     }
 
     size_t const rb_3size(msg_size*3);
@@ -336,38 +340,38 @@ START_TEST(recovery)
     {
         rb_ctx ctx(rb_3size);
 
-        ck_assert_msg(ctx.rb.size() == ctx.size,
-                      "Expected %zd, got %zd", ctx.size, ctx.rb.size());
+        fail_if (ctx.rb.size() != ctx.size,
+                 "Expected %zd, got %zd", ctx.size, ctx.rb.size());
 
         if (ctx.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(!ctx.s2p.empty());
-        ck_assert(ctx.s2p.size() == 2);
-        ck_assert(ctx.s2p.index_begin() != seqno_min);
-        ck_assert(ctx.s2p.index_back()  == seqno_max);
+        fail_if(ctx.s2p.empty());
+        fail_if(ctx.s2p.size() != 2);
+        fail_if(ctx.s2p.begin()->first  == seqno_min);
+        fail_if(ctx.s2p.rbegin()->first != seqno_max);
 
         void* m(ctx.add_msg(msgs[8]));
-        ck_assert(NULL != m);
-        ck_assert(*ctx.s2p.find(msgs[8].g) == m);
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[8].g)->second != m);
 
         m = ctx.add_msg(msgs[9]);
-        ck_assert(NULL != m);
-        ck_assert(*ctx.s2p.find(msgs[9].g) == m);
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[9].g)->second != m);
 
         m = ctx.add_msg(msgs[7]);
-        ck_assert(NULL != m);
-        ck_assert(msgs[7].g <= 0);
-        ck_assert(ctx.s2p.find(msgs[7].g) == ctx.s2p.end());
+        fail_if (NULL == m);
+        fail_if (msgs[7].g > 0);
+        fail_if (ctx.s2p.find(msgs[7].g) != ctx.s2p.end());
         // here we should have rollover
-        ck_assert(ptr2BH(m) == BH_cast(ctx.rb.start()));
+        fail_if (ptr2BH(m) != BH_cast(ctx.rb.start()));
 
-        seqno_min = ctx.s2p.index_front();
-        seqno_max = ctx.s2p.index_back();
+        seqno_min = ctx.s2p.begin()->first;
+        seqno_max = ctx.s2p.rbegin()->first;
     }
 
     /* what we should have now is |***---777| - only one segment, at the end */
@@ -375,80 +379,80 @@ START_TEST(recovery)
         /* first open this with known offset */
         rb_ctx ctx0(rb_3size);
 
-        ck_assert_msg(ctx0.rb.size() == ctx0.size,
-                      "Expected %zd, got %zd", ctx0.size, ctx0.rb.size());
+        fail_if (ctx0.rb.size() != ctx0.size,
+                 "Expected %zd, got %zd", ctx0.size, ctx0.rb.size());
 
         if (ctx0.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx0.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(!ctx0.s2p.empty());
-        ck_assert(ctx0.s2p.size() == 1);
-        ck_assert(ctx0.s2p.index_front() == seqno_max);
-        ck_assert(ctx0.s2p.index_back()  == seqno_max);
+        fail_if(ctx0.s2p.empty());
+        fail_if(ctx0.s2p.size() != 1);
+        fail_if(ctx0.s2p.begin()->first  != seqno_max);
+        fail_if(ctx0.s2p.rbegin()->first != seqno_max);
 
         /* now try to open unclosed file. Results should be the same */
         rb_ctx ctx(rb_3size);
 
-        ck_assert_msg(ctx.rb.size() == ctx.size,
-                      "Expected %zd, got %zd", ctx.size, ctx.rb.size());
+        fail_if (ctx.rb.size() != ctx.size,
+                 "Expected %zd, got %zd", ctx.size, ctx.rb.size());
 
         if (ctx.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(!ctx.s2p.empty());
-        ck_assert(ctx.s2p.size() == 1);
-        ck_assert(ctx.s2p.index_front() == seqno_max);
-        ck_assert(ctx.s2p.index_back()  == seqno_max);
+        fail_if(ctx.s2p.empty());
+        fail_if(ctx.s2p.size() != 1);
+        fail_if(ctx.s2p.begin()->first  != seqno_max);
+        fail_if(ctx.s2p.rbegin()->first != seqno_max);
 
-        ck_assert(seqno_max >= 1);
-        ck_assert(seqno_min == seqno_max);
+        fail_if(seqno_max < 1);
+        fail_if(seqno_min != seqno_max);
     }
 
     ::unlink(RB_NAME.c_str());
 
-    /* test for single segment in the middle */
-    ptrdiff_t third_buffer_offset(0);
+    /* test for singe segment in the middle */
+    void* third_buffer(NULL);
     {
         rb_ctx ctx(rb_3size, false);
 
-        ck_assert_msg(ctx.rb.size() == ctx.size,
-                      "Expected %zd, got %zd", ctx.size, ctx.rb.size());
+        fail_if (ctx.rb.size() != ctx.size,
+                 "Expected %zd, got %zd", ctx.size, ctx.rb.size());
 
         if (ctx.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(ctx.s2p.empty());
+        fail_if(!ctx.s2p.empty());
 
         void* m(ctx.add_msg(msgs[3]));
-        ck_assert(NULL != m);
-        ck_assert(ctx.s2p.find(msgs[3].g) == ctx.s2p.end());
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[3].g) != ctx.s2p.end());
 
         m = ctx.add_msg(msgs[4]);
-        ck_assert(NULL != m);
-        ck_assert(*ctx.s2p.find(msgs[4].g) == m);
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[4].g)->second != m);
 
         m = ctx.add_msg(msgs[5]);
-        ck_assert(NULL != m);
-        ck_assert(ctx.s2p.find(msgs[5].g) == ctx.s2p.end());
-        third_buffer_offset = ctx.rb.offset(m);
+        fail_if (NULL == m);
+        fail_if (ctx.s2p.find(msgs[5].g) != ctx.s2p.end());
+        third_buffer = m;
 
-        ck_assert(!ctx.s2p.empty());
-        ck_assert(ctx.s2p.size() == 1);
-        seqno_min = ctx.s2p.index_front();
-        seqno_max = ctx.s2p.index_back();
-        ck_assert(seqno_min == seqno_max);
+        fail_if(ctx.s2p.empty());
+        fail_if(ctx.s2p.size() != 1);
+        seqno_min = ctx.s2p.begin()->first;
+        seqno_max = ctx.s2p.rbegin()->first;
+        fail_if(seqno_min != seqno_max);
     }
 
     /* now the situation should be |***444***| - only one segment, in the middle,
@@ -456,50 +460,48 @@ START_TEST(recovery)
     {
         rb_ctx ctx(rb_3size);
 
-        ck_assert_msg(ctx.rb.size() == ctx.size,
-                      "Expected %zd, got %zd", ctx.size, ctx.rb.size());
+        fail_if (ctx.rb.size() != ctx.size,
+                 "Expected %zd, got %zd", ctx.size, ctx.rb.size());
 
         if (ctx.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(!ctx.s2p.empty());
-        ck_assert(ctx.s2p.size() == 1);
-        ck_assert(seqno_min == ctx.s2p.index_begin());
-        ck_assert(seqno_max == ctx.s2p.index_back());
-        ck_assert(seqno_min == seqno_max);
+        fail_if(ctx.s2p.empty());
+        fail_if(ctx.s2p.size() != 1);
+        fail_if(seqno_min != ctx.s2p.begin()->first);
+        fail_if(seqno_max != ctx.s2p.rbegin()->first);
+        fail_if(seqno_min != seqno_max);
     }
 
-    /* now the situation should be |---444---| - only one segment,in the middle,
+    /* now the situation should be |---444---| - only one segment, in the middle,
      * reopen the file a second time - to trigger a rollover bug */
     {
         rb_ctx ctx(rb_3size);
 
-        ck_assert_msg(ctx.rb.size() == ctx.size,
-                      "Expected %zd, got %zd", ctx.size, ctx.rb.size());
+        fail_if (ctx.rb.size() != ctx.size,
+                 "Expected %zd, got %zd", ctx.size, ctx.rb.size());
 
         if (ctx.gid != GID)
         {
             std::ostringstream os;
             os << "Expected GID: " << GID << ", got: " << ctx.gid;
-            ck_abort_msg("%s", os.str().c_str());
+            fail(os.str().c_str());
         }
 
-        ck_assert(!ctx.s2p.empty());
-        ck_assert(ctx.s2p.size() == 1);
-        ck_assert(seqno_min == ctx.s2p.index_front());
-        ck_assert(seqno_max == ctx.s2p.index_back());
-        ck_assert(seqno_min == seqno_max);
+        fail_if(ctx.s2p.empty());
+        fail_if(ctx.s2p.size() != 1);
+        fail_if(seqno_min != ctx.s2p.begin()->first);
+        fail_if(seqno_max != ctx.s2p.rbegin()->first);
+        fail_if(seqno_min != seqno_max);
 
         // must be allocated right after the recovered buffer
         void* m(ctx.add_msg(msgs[3]));
-        ck_assert(NULL != m);
-        ck_assert_msg(third_buffer_offset == ctx.rb.offset(m),
-                      "expected %zd, got %zd",
-                      third_buffer_offset, ctx.rb.offset(m));
+        fail_if (NULL == m);
+        fail_if (third_buffer != m);
     }
 
     ::unlink(RB_NAME.c_str());
