@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010 Codership Oy <info@codership.com>
+ * Copyright (C) 2010-2017 Codership Oy <info@codership.com>
  */
 
 #ifndef GCOMM_ASIO_TCP_HPP
@@ -8,21 +8,23 @@
 #include "socket.hpp"
 #include "asio_protonet.hpp"
 
+#include "gu_array.hpp"
+#include "gu_shared_ptr.hpp"
+
 #include <boost/bind.hpp>
-#include <boost/array.hpp>
-#include <boost/enable_shared_from_this.hpp>
 #include <vector>
 #include <deque>
 
 //
-// Boost enable_shared_from_this<> does not have virtual destructor,
-// therefore need to ignore -Weffc++
+// Boost and std:: enable_shared_from_this<> does not have virtual destructor,
+// therefore need to ignore -Weffc++ and -Wnon-virtual-dtor
 //
 #if defined(__GNUG__)
 # if (__GNUC__ == 4 && __GNUC_MINOR__ >= 6) || (__GNUC__ > 4)
 #  pragma GCC diagnostic push
 # endif // (__GNUC__ == 4 && __GNUC_MINOR__ >= 6) || (__GNUC__ > 4)
 # pragma GCC diagnostic ignored "-Weffc++"
+# pragma GCC diagnostic ignored "-Wnon-virtual-dtor"
 #endif
 
 namespace gcomm
@@ -36,15 +38,13 @@ namespace gcomm
 
 class gcomm::AsioTcpSocket :
     public gcomm::Socket,
-    public boost::enable_shared_from_this<AsioTcpSocket>
+    public gu::enable_shared_from_this<AsioTcpSocket>::type
 {
 public:
     AsioTcpSocket(AsioProtonet& net, const gu::URI& uri);
     ~AsioTcpSocket();
     void failed_handler(const asio::error_code& ec, const std::string& func, int line);
-#ifdef HAVE_ASIO_SSL_HPP
     void handshake_handler(const asio::error_code& ec);
-#endif // HAVE_ASIO_SSL_HPP
     void connect_handler(const asio::error_code& ec);
     void connect(const gu::URI& uri);
     void close();
@@ -71,8 +71,8 @@ private:
     void operator=(const AsioTcpSocket&);
 
     void set_socket_options();
-    void read_one(boost::array<asio::mutable_buffer, 1>& mbs);
-    void write_one(const boost::array<asio::const_buffer, 2>& cbs);
+    void read_one(gu::array<asio::mutable_buffer, 1>::type& mbs);
+    void write_one(const gu::array<asio::const_buffer, 2>::type& cbs);
     void close_socket();
 
     // call to assign local/remote addresses at the point where it
@@ -89,9 +89,7 @@ private:
 
     AsioProtonet&                             net_;
     asio::ip::tcp::socket                     socket_;
-#ifdef HAVE_ASIO_SSL_HPP
     asio::ssl::stream<asio::ip::tcp::socket>* ssl_socket_;
-#endif // HAVE_ASIO_SSL_HPP
     std::deque<Datagram>                      send_q_;
     std::vector<gu::byte_t>                   recv_buf_;
     size_t                                    recv_offset_;
@@ -121,9 +119,6 @@ public:
 
     AsioTcpAcceptor(AsioProtonet& net, const gu::URI& uri);
     ~AsioTcpAcceptor();
-    void accept_handler(
-        SocketPtr socket,
-        const asio::error_code& error);
     void listen(const gu::URI& uri);
     std::string listen_addr() const;
     void close();
@@ -137,6 +132,10 @@ public:
     SocketId id() const { return &acceptor_; }
 
 private:
+    void accept_handler(
+        SocketPtr socket,
+        const asio::error_code& error);
+
     AsioProtonet& net_;
     asio::ip::tcp::acceptor acceptor_;
     SocketPtr accepted_socket_;
