@@ -1,9 +1,8 @@
 /*
- * Copyright (C) 2012-2019 Codership Oy <info@codership.com>
+ * Copyright (C) 2012-2014 Codership Oy <info@codership.com>
  */
 
 #include "defaults.hpp"
-#include "asio_tcp.hpp"
 
 #include "gcomm/common.hpp"
 
@@ -17,10 +16,7 @@ namespace gcomm
 
     std::string const Defaults::ProtonetVersion         = "0";
     std::string const Defaults::SocketChecksum          = "2";
-    std::string const Defaults::SocketRecvBufSize       =
-        GCOMM_ASIO_AUTO_BUF_SIZE;
-    std::string const Defaults::SocketSendBufSize       =
-        GCOMM_ASIO_AUTO_BUF_SIZE;
+    std::string const Defaults::SocketRecvBufSize       = "212992";
     std::string const Defaults::GMCastVersion           = "0";
     std::string const Defaults::GMCastTcpPort           = BASE_PORT_DEFAULT;
     std::string const Defaults::GMCastSegment           = "0";
@@ -46,6 +42,7 @@ namespace gcomm
     std::string const Defaults::EvsDelayMargin          = "PT1S";
     std::string const Defaults::EvsDelayedKeepPeriod    = "PT30S";
     std::string const Defaults::EvsAutoEvict            = "0";
+    std::string const Defaults::EvsVersion              = "1";
     std::string const Defaults::PcAnnounceTimeout       = "PT3S";
     std::string const Defaults::PcChecksum              = "false";
     std::string const Defaults::PcIgnoreQuorum          = "false";
