@@ -1,5 +1,7 @@
 /*
- * Copyright (C) 2009-2019 Codership Oy <info@codership.com>
+ * Copyright (C) 2009 Codership Oy <info@codership.com>
+ *
+ * $Id$
  */
 
 #ifndef EVS_MESSAGE2_HPP
@@ -76,6 +78,8 @@ public:
         im_range_    (mn.im_range_   )
     { }
 
+    MessageNode& operator=(const MessageNode& other) = default;
+
     bool          operational() const { return operational_       ; }
     bool          suspected()   const { return suspected_         ; }
     bool          evicted()     const { return evicted_           ; }
@@ -369,7 +373,7 @@ public:
         install_view_id_ (install_view_id),
         range_uuid_      (range_uuid),
         range_           (range),
-        tstamp_          (gu::datetime::Date::monotonic()),
+        tstamp_          (gu::datetime::Date::now()),
         node_list_       (node_list),
         delayed_list_    ()
     { }
