@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2020 Codership Oy <info@codership.com>
+ * Copyright (C) 2009-2014 Codership Oy <info@codership.com>
  */
 
 #include "gcomm/util.hpp"
@@ -7,6 +7,10 @@
 #include "gcomm/datagram.hpp"
 #include "gcomm/conf.hpp"
 
+#ifdef HAVE_ASIO_HPP
+#include "asio_protonet.hpp"
+#endif // HAVE_ASIO_HPP
+
 #include "check_gcomm.hpp"
 
 #include "gu_logger.hpp"
@@ -31,26 +35,26 @@ START_TEST(test_datagram)
 
     // Header check
     gcomm::NetHeader hdr(42, 0);
-    ck_assert(hdr.len() == 42);
-    ck_assert(hdr.has_crc32() == false);
-    ck_assert(hdr.version() == 0);
+    fail_unless(hdr.len() == 42);
+    fail_unless(hdr.has_crc32() == false);
+    fail_unless(hdr.version() == 0);
 
     hdr.set_crc32(1234, NetHeader::CS_CRC32);
-    ck_assert(hdr.has_crc32() == true);
-    ck_assert(hdr.len() == 42);
+    fail_unless(hdr.has_crc32() == true);
+    fail_unless(hdr.len() == 42);
 
     gcomm::NetHeader hdr1(42, 1);
-    ck_assert(hdr1.len() == 42);
-    ck_assert(hdr1.has_crc32() == false);
-    ck_assert(hdr1.version() == 1);
+    fail_unless(hdr1.len() == 42);
+    fail_unless(hdr1.has_crc32() == false);
+    fail_unless(hdr1.version() == 1);
 
     gu::byte_t hdrbuf[NetHeader::serial_size_];
-    ck_assert(serialize(hdr1, hdrbuf, sizeof(hdrbuf), 0) ==
+    fail_unless(serialize(hdr1, hdrbuf, sizeof(hdrbuf), 0) ==
                 NetHeader::serial_size_);
     try
     {
         unserialize(hdrbuf, sizeof(hdrbuf), 0, hdr);
-        ck_abort();
+        fail("");
     }
     catch (Exception& e)
     {
@@ -66,23 +70,23 @@ START_TEST(test_datagram)
     gu::Buffer buf(b, b + sizeof(b));
 
     gcomm::Datagram dg(buf);
-    ck_assert(dg.len() == sizeof(b));
+    fail_unless(dg.len() == sizeof(b));
 
     // Normal copy construction
     gcomm::Datagram dgcopy(buf);
-    ck_assert(dgcopy.len() == sizeof(b));
-    ck_assert(memcmp(dgcopy.header() + dgcopy.header_offset(),
+    fail_unless(dgcopy.len() == sizeof(b));
+    fail_unless(memcmp(dgcopy.header() + dgcopy.header_offset(),
                        dg.header() + dg.header_offset(),
                        dg.header_len()) == 0);
-    ck_assert(dgcopy.payload() == dg.payload());
+    fail_unless(dgcopy.payload() == dg.payload());
 
     // Copy construction from offset of 16
     gcomm::Datagram dg16(dg, 16);
     log_info << dg16.len();
-    ck_assert(dg16.len() - dg16.offset() == sizeof(b) - 16);
+    fail_unless(dg16.len() - dg16.offset() == sizeof(b) - 16);
     for (gu::byte_t i = 0; i < sizeof(b) - 16; ++i)
     {
-        ck_assert(dg16.payload()[i + dg16.offset()] == i + 16);
+        fail_unless(dg16.payload()[i + dg16.offset()] == i + 16);
     }
 
 #if 0
@@ -91,24 +95,24 @@ START_TEST(test_datagram)
     // dg
     dg16.normalize();
 
-    ck_assert(dg16.len() == sizeof(b) - 16);
+    fail_unless(dg16.len() == sizeof(b) - 16);
     for (byte_t i = 0; i < sizeof(b) - 16; ++i)
     {
-        ck_assert(dg16.payload()[i] == i + 16);
+        fail_unless(dg16.payload()[i] == i + 16);
     }
 
-    ck_assert(dg.len() == sizeof(b));
+    fail_unless(dg.len() == sizeof(b));
     for (byte_t i = 0; i < sizeof(b); ++i)
     {
-        ck_assert(dg.payload()[i] == i);
+        fail_unless(dg.payload()[i] == i);
     }
 
     Datagram dgoff(buf, 16);
     dgoff.header().resize(8);
     dgoff.set_header_offset(4);
-    ck_assert(dgoff.len() == buf.size() + 4);
-    ck_assert(dgoff.header_offset() == 4);
-    ck_assert(dgoff.header().size() == 8);
+    fail_unless(dgoff.len() == buf.size() + 4);
+    fail_unless(dgoff.header_offset() == 4);
+    fail_unless(dgoff.header().size() == 8);
     for (byte_t i = 0; i < 4; ++i)
     {
         *(&dgoff.header()[0] + i) = i;
@@ -116,15 +120,62 @@ START_TEST(test_datagram)
 
     dgoff.normalize();
 
-    ck_assert(dgoff.len() == sizeof(b) - 16 + 4);
-    ck_assert(dgoff.header_offset() == 0);
-    ck_assert(dgoff.header().size() == 0);
+    fail_unless(dgoff.len() == sizeof(b) - 16 + 4);
+    fail_unless(dgoff.header_offset() == 0);
+    fail_unless(dgoff.header().size() == 0);
 #endif // 0
 }
 END_TEST
 
 
+#if defined(HAVE_ASIO_HPP)
+START_TEST(test_asio)
+{
+    gu::Config conf;
+    gu::ssl_register_params(conf);
+    gcomm::Conf::register_params(conf);
+    AsioProtonet pn(conf);
+    string uri_str("tcp://127.0.0.1:0");
 
+    Acceptor* acc = pn.acceptor(uri_str);
+    acc->listen(uri_str);
+    uri_str = acc->listen_addr();
+
+    SocketPtr cl = pn.socket(uri_str);
+    cl->connect(uri_str);
+    pn.event_loop(gu::datetime::Sec);
+
+    SocketPtr sr = acc->accept();
+    fail_unless(sr->state() == Socket::S_CONNECTED);
+
+    vector<byte_t> buf(cl->mtu());
+    for (size_t i = 0; i < buf.size(); ++i)
+    {
+        buf[i] = static_cast<byte_t>(i & 0xff);
+    }
+
+    for (size_t i = 0; i < 13; ++i)
+    {
+        Datagram dg(Buffer(&buf[0], &buf[0] + buf.size()));
+        cl->send(dg);
+    }
+    pn.event_loop(gu::datetime::Sec);
+
+    delete acc;
+
+}
+END_TEST
+#endif // HAVE_ASIO_HPP
+
+START_TEST(test_protonet)
+{
+    gu::Config conf;
+    gu::ssl_register_params(conf);
+    gcomm::Conf::register_params(conf);
+    Protonet* pn(Protonet::create(conf));
+    pn->event_loop(1);
+}
+END_TEST
 
 START_TEST(test_view_state)
 {
@@ -145,7 +196,7 @@ START_TEST(test_view_state)
         std::istringstream is(os.str());
         view2.read_stream(is);
 
-        ck_assert(view == view2);
+        fail_unless(view == view2);
     }
 
     // Create configuration to set file name.
@@ -165,7 +216,7 @@ START_TEST(test_view_state)
         std::istringstream is(os.str());
         vst2.read_stream(is);
 
-        ck_assert(vst == vst2);
+        fail_unless(vst == vst2);
     }
 
     // test write file and read file.
@@ -174,7 +225,7 @@ START_TEST(test_view_state)
     View view_3;
     ViewState vst3(my_uuid_3, view_3, conf);
     vst3.read_file();
-    ck_assert(vst == vst3);
+    fail_unless(vst == vst3);
     ViewState::remove_file(conf);
 }
 END_TEST
@@ -189,6 +240,16 @@ Suite* util_suite()
     tcase_add_test(tc, test_datagram);
     suite_add_tcase(s, tc);
 
+#ifdef HAVE_ASIO_HPP
+    tc = tcase_create("test_asio");
+    tcase_add_test(tc, test_asio);
+    suite_add_tcase(s, tc);
+#endif // HAVE_ASIO_HPP
+
+    tc = tcase_create("test_protonet");
+    tcase_add_test(tc, test_protonet);
+    suite_add_tcase(s, tc);
+
     tc = tcase_create("test_view_state");
     tcase_add_test(tc, test_view_state);
     suite_add_tcase(s, tc);
