@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2020 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2017 Codership Oy <info@codership.com>
  *
  * $Id$
  */
@@ -16,7 +16,8 @@
 #include <gu_config.h>
 #include <gu_buf.h>
 #include <gu_errno.h>
-#include <gu_uuid.h>
+#include <gu_uuid.hpp>
+#include <gu_gtid.hpp>
 #include <gu_status.hpp>
 
 #include <stdint.h>
@@ -24,6 +25,9 @@
 #include <unistd.h>
 #include <stdio.h>
 
+#include <vector>
+#include <string>
+
 /*! @typedef @brief Sequence number type. */
 typedef int64_t gcs_seqno_t;
 
@@ -61,22 +65,20 @@ gcs_create  (gu_config_t* conf, gcache_t* cache,
  * Serves to provide group history persistence after process restart (in case
  * these data were saved somewhere on persistent storage or the like). If these
  * values are provided, it is only a hint for the group, as they might be
- * outdated. Actual seqno and UUID are returned in GCS_ACT_CONF action (see
+ * outdated. Actual seqno and UUID are returned in GCS_ACT_CCHANGE action (see
  * below) and are determined by quorum.
  *
  * This function must be called before gcs_open() or after gcs_close().
  *
- * @param seqno Sequence number of the application state (last action applied).
- *              Should be negative for undefined state.
- * @param uuid  UUID of the sequence (group ID).
- *              Should be all zeroes for undefined state.
+ * @param position Global Transaction ID corresponding to the current
+ *                 application state.
+ *                 Should be undefined for undefined state.
  *
  * @return 0 in case of success, -EBUSY if conneciton is already opened,
  *         -EBADFD if connection object is being destroyed.
  */
-extern long gcs_init (gcs_conn_t   *conn,
-                      gcs_seqno_t   seqno,
-                      const uint8_t uuid[GCS_UUID_LEN]);
+extern long gcs_init (gcs_conn_t*     conn,
+                      const gu::GTID& position);
 
 /*! @brief Opens connection to group (joins channel).
  *
@@ -138,28 +140,30 @@ extern long gcs_wait (gcs_conn_t *conn);
  *
  * The purpose of GCS library is to hide message handling from application.
  * Therefore application deals only with "actions".
- * Application can only send actions of types GCS_ACT_TORDERED,
+ * Application can only send actions of types GCS_ACT_WRITESET,
  * GCS_ACT_COMMIT_CUT and GCS_ACT_STATE_REQ.
- * Actions of type GCS_ACT_SYNC, GCS_ACT_CONF are generated by the library.
+ * Actions of type GCS_ACT_SYNC, GCS_ACT_CCHANGE are generated by the library.
  */
 typedef enum gcs_act_type
 {
 /* ordered actions */
-    GCS_ACT_TORDERED,   //! action representing state change, will be assigned
+    GCS_ACT_WRITESET,   //! action representing state change, will be assigned
                         //  global seqno
     GCS_ACT_COMMIT_CUT, //! group-wide action commit cut
     GCS_ACT_STATE_REQ,  //! request for state transfer
-    GCS_ACT_CONF,       //! new configuration
+    GCS_ACT_CCHANGE,    //! group configuration change
     GCS_ACT_JOIN,       //! joined group (received all state data)
     GCS_ACT_SYNC,       //! synchronized with group
     GCS_ACT_FLOW,       //! flow control
+    GCS_ACT_VOTE,       //! vote on GTID outcome
     GCS_ACT_SERVICE,    //! service action, sent by GCS
     GCS_ACT_ERROR,      //! error happened while receiving the action
-    GCS_ACT_INCONSISTENCY,//! inconsistency event
     GCS_ACT_UNKNOWN     //! undefined/unknown action type
 }
 gcs_act_type_t;
 
+#define GCS_VOTE_REQUEST 1 /* vote request indicator */
+
 /*! String representations of action types */
 extern const char* gcs_act_type_to_str(gcs_act_type_t);
 
@@ -173,6 +177,7 @@ extern const char* gcs_act_type_to_str(gcs_act_type_t);
  * @param act_size   total action size (the sum of buffer sizes)
  * @param act_type   action type
  * @param scheduled  whether the call was scheduled by gcs_schedule()
+ * @param grab       use gcs_sm_grab() instead of gcs_sm_enter()
  * @return           negative error code, action size in case of success
  * @retval -EINTR    thread was interrupted while waiting to enter the monitor
  */
@@ -180,7 +185,8 @@ extern long gcs_sendv (gcs_conn_t*          conn,
                        const struct gu_buf* act_bufs,
                        size_t               act_size,
                        gcs_act_type_t       act_type,
-                       bool                 scheduled);
+                       bool                 scheduled,
+                       bool                 grab);
 
 /*! A wrapper for single buffer communication */
 static inline long gcs_send (gcs_conn_t*    const conn,
@@ -190,18 +196,20 @@ static inline long gcs_send (gcs_conn_t*    const conn,
                              bool           const scheduled)
 {
     struct gu_buf const buf = { act, static_cast<ssize_t>(act_size) };
-    return gcs_sendv (conn, &buf, act_size, act_type, scheduled);
+    return gcs_sendv (conn, &buf, act_size, act_type, scheduled, false);
 }
 
 /*!*/
 struct gcs_action {
-    const void*    buf; /*! unlike input, output goes as a single buffer */
-    ssize_t        size;
     gcs_seqno_t    seqno_g;
     gcs_seqno_t    seqno_l;
+    const void*    buf; /*! unlike input, output goes as a single buffer */
+    int32_t        size;
     gcs_act_type_t type;
 };
 
+std::ostream& operator <<(std::ostream& os, const gcs_action& act);
+
 /*! @brief Replicates a vector of buffers as a single action.
  * Sends action to group and blocks until it is received. Upon return global
  * and local IDs are set. Arguments are the same as in gcs_recv().
@@ -278,14 +286,14 @@ extern long gcs_interrupt (gcs_conn_t* conn, long handle);
 extern long gcs_resume_recv (gcs_conn_t* conn);
 
 /*!
- * After action with this seqno is applied, this thread is guaranteed to see
+ * After action with this gtid is applied, this thread is guaranteed to see
  * all the changes made by the client, even on other nodes.
  *
  * @retval 0       success
  * @retval -EPERM  operation not permitted (in NON_PRIMARY state)
  * @retval -EAGAIN operation may be retried later (in transient state)
  */
-extern long gcs_caused (gcs_conn_t* conn, gcs_seqno_t& seqno);
+extern long gcs_caused (gcs_conn_t* conn, gu::GTID& gtid);
 
 /*! @brief Sends state transfer request
  * Broadcasts state transfer request which will be passed to one of the
@@ -298,41 +306,41 @@ extern long gcs_caused (gcs_conn_t* conn, gcs_seqno_t& seqno);
  * @param size  request size
  * @param donor desired state transfer donor name. Supply empty string to
  *              choose automatically.
- * @param seqno response to request was ordered with this seqno.
+ * @param ist_gtid where to start IST from
+ * @param order response to request was ordered with this local order.
  *              Must be skipped in local queues.
  * @return negative error code, index of state transfer donor in case of success
  *         (notably, -EAGAIN means try later, -EHOSTUNREACH means desired donor
  *         is unavailable)
  */
-extern long gcs_request_state_transfer (gcs_conn_t  *conn,
-                                        int          ver,
-                                        const void  *req,
-                                        size_t       size,
-                                        const char  *donor,
-                                        const gu_uuid_t* ist_uuid,
-                                        gcs_seqno_t ist_seqno,
-                                        gcs_seqno_t *seqno);
+extern long gcs_request_state_transfer (gcs_conn_t*     conn,
+                                        int             ver,
+                                        const void*     req,
+                                        size_t          size,
+                                        const char*     donor,
+                                        const gu::GTID& ist_gtid,
+                                        gcs_seqno_t&    order);
 
 /*! @brief Turns off flow control on the node.
  * Effectively desynchronizes the node from the cluster (while the node keeps on
  * receiving all the actions). Requires gcs_join() to return to normal.
  *
  * @param conn  connection to group
- * @param seqno response to request was ordered with this seqno.
+ * @param order response to request was ordered with this seqno.
  *              Must be skipped in local queues.
  * @return negative error code, 0 in case of success.
  */
-extern long gcs_desync (gcs_conn_t* conn, gcs_seqno_t* seqno);
+extern long gcs_desync (gcs_conn_t* conn, gcs_seqno_t& order);
 
 /*! @brief Informs group on behalf of donor that state stransfer is over.
  * If status is non-negative, joiner will be considered fully joined to group.
  *
  * @param conn opened connection to group
- * @param status negative error code in case of state transfer failure,
- *               0 or (optional) seqno corresponding to transferred state.
+ * @param gtid containing negative error code in case of state transfer failure,
+ *             or gtid of joined state.
  * @return negative error code, 0 in case of success
  */
-extern long gcs_join (gcs_conn_t *conn, gcs_seqno_t status);
+extern long gcs_join (gcs_conn_t *conn, const gu::GTID& gtid, int code);
 
 /*! @brief Allocate local seqno for accessing local resources.
  *
@@ -348,7 +356,26 @@ extern gcs_seqno_t gcs_local_sequence(gcs_conn_t* conn);
 /* Service functions */
 
 /*! Informs group about the last applied action on this node */
-extern long gcs_set_last_applied (gcs_conn_t* conn, gcs_seqno_t seqno);
+extern long
+gcs_set_last_applied (gcs_conn_t* conn, const gu::GTID& gtid);
+
+/*! @return currently established GCS protocol */
+extern int
+gcs_proto_ver(gcs_conn_t* conn);
+
+/*! @brief Vote on the error code that resulted from processing the gtid action.
+ *
+ * Blocks until consensus is reached or call fails.
+ *
+ * @param msg optional error message (should not be node-specific)
+ * @param msg_len message length
+ *
+ * @return 0 for majority agrees on error, 1 for majority disagrees with error,
+ *         negative errno for technical call failure.
+ */
+extern int
+gcs_vote (gcs_conn_t* conn, const gu::GTID& gtid, uint64_t code,
+          const void* msg, size_t msg_len);
 
 /* GCS Configuration */
 
@@ -406,19 +433,46 @@ gcs_node_state_t;
 extern const char*
 gcs_node_state_to_str (gcs_node_state_t state);
 
-/*! New configuration action */
-typedef struct gcs_act_conf {
-    gcs_seqno_t      seqno;    //! last global seqno applied by this group
-    gcs_seqno_t      conf_id;  //! configuration ID (-1 if non-primary)
-    uint8_t          uuid[GCS_UUID_LEN];/// group UUID
-    long             memb_num; //! number of members in configuration
-    long             my_idx;   //! index of this node in the configuration
-    gcs_node_state_t my_state; //! current node state
-    int              repl_proto_ver; //! replicator  protocol version to use
-    int              appl_proto_ver; //! application protocol version to use
-    char             data[1];  /*! member array (null-terminated ID, name,
-                                *  incoming address, 8-byte cached seqno) */
-} gcs_act_conf_t;
+/*! New configuration action deserialized */
+struct gcs_act_cchange
+{
+    gcs_act_cchange();
+    gcs_act_cchange(const void* buf, int size);
+
+    int write(void** buf) const;
+    bool operator==(const gcs_act_cchange& other) const;
+
+    struct member
+    {
+        member() : uuid_(), name_(), incoming_(), cached_(), state_() {}
+        bool operator==(const member& other) const;
+        gu_uuid_t      uuid_;
+        std::string    name_;
+        std::string    incoming_;
+        gcs_seqno_t    cached_;
+        gcs_node_state state_;
+    };
+
+    std::vector<member> memb;
+    gu_uuid_t           uuid;     //! group UUID
+    gcs_seqno_t         seqno;    //! last global seqno applied by this group
+    gcs_seqno_t         conf_id;  //! configuration ID (-1 if non-primary)
+    gcs_seqno_t         vote_seqno;
+    int64_t             vote_res;
+    int                 repl_proto_ver; //! replicator  protocol version to use
+    int                 appl_proto_ver; //! application protocol version to use
+};
+
+std::ostream&
+operator <<(std::ostream& os, const struct gcs_act_cchange& cc);
+
+struct gcs_backend_stats {
+    struct stats_t {
+        const char* key;
+        const char* value;
+    }* stats;
+    void* ctx;
+};
 
 struct gcs_stats
 {
@@ -436,8 +490,7 @@ struct gcs_stats
     int       send_q_len;     //! current send queue length
     int       send_q_len_max; //! maximum send queue length
     int       send_q_len_min; //! minimum send queue length
-    bool      fc_active;      //! flow control is currently active
-    bool      fc_requested;   //! flow control is requested by this node
+    struct gcs_backend_stats backend_stats; //! backend stats.
 };
 
 /*! Fills stats struct */
