@@ -4,6 +4,9 @@ set -eu
 
 # $Id$
 
+# Galera library version
+VERSION="26.4.3"
+
 get_cores()
 {
     case $OS in
@@ -35,15 +38,14 @@ Options:
     --dl            set debug level for Scons build (1, implies -c)
     -r|--release    release number
     -m32/-m64       build 32/64-bit binaries for x86
-    -p|--package    build RPM packages at the end (DEB not supported).
+    -p|--package    build RPM and DEB packages at the end.
     --with-spread   configure build with spread backend (implies -c to gcs)
     --source        build source packages
     --sb            skip actual build, use the existing binaries
-    --cmake         build using CMake build system (yes)
-    --co            CMake options
-    --scons         build using Scons build system (no)
+    --scons         build using Scons build system (yes)
     --so            Sconscript option
     -j|--jobs       how many parallel jobs to use for Scons (1)
+    -a|--archive    Generate source archive (tar.gz)
     "\nSet DISABLE_GCOMM/DISABLE_VSBES to 'yes' to disable respective modules"
 EOF
 }
@@ -58,9 +60,7 @@ RELEASE=${RELEASE:-""}
 SOURCE=${SOURCE:-"no"}
 DEBUG=${DEBUG:-"no"}
 DEBUG_LEVEL=${DEBUG_LEVEL:-"0"}
-CMAKE=${CMAKE:-"yes"}
-CMAKE_OPTS=${CMAKE_OPTS:-""}
-SCONS=${SCONS:-"no"}
+SCONS=${SCONS:-"yes"}
 SCONS_OPTS=${SCONS_OPTS:-""}
 export JOBS=${JOBS:-"$(get_cores)"}
 SCRATCH=${SCRATCH:-"no"}
@@ -68,6 +68,8 @@ OPT="yes"
 NO_STRIP=${NO_STRIP:-"no"}
 WITH_SPREAD="no"
 RUN_TESTS=${RUN_TESTS:-1}
+ARCHIVE=${ARCHIVE:-"no"}
+
 if [ "$OS" == "FreeBSD" ]; then
   chown=/usr/sbin/chown
   true=/usr/bin/true
@@ -92,9 +94,12 @@ elif [ "$OS" == "FreeBSD" ]; then
   EXTRA_SYSROOT=/usr/local
 fi
 
+which dpkg >/dev/null 2>&1 && DEBIAN=${DEBIAN:-1} || DEBIAN=${DEBIAN:-0}
+
 if [ "$OS" == "FreeBSD" ]; then
-    CC=${CC:-"clang"}
-    CXX=${CXX:-"clang++"}
+    CC=${CC:-"gcc48"}
+    CXX=${CXX:-"g++48"}
+    LD_LIBRARY_PATH=${LD_LIBRARY_PATH:-"/usr/local/lib/$(basename $CC)"}
 else
     CC=${CC:-"gcc"}
     CXX=${CXX:-"g++"}
@@ -105,6 +110,20 @@ then
     echo "$CC"  | grep "ccache" > /dev/null || CC="ccache $CC"
     echo "$CXX" | grep "ccache" > /dev/null || CXX="ccache $CXX"
 fi
+
+# MariaDB Centos5 & SLES-SP1 buildbot images require LD_LIBRARY_PATH to
+# /usr/local/lib
+if [ -r /etc/redhat-release ]; then
+  if grep 'CentOS release 5' /etc/redhat-release > /dev/null 2>&1; then
+    LD_LIBRARY_PATH=/usr/local/lib64/:/usr/local/lib/:${LD_LIBRARY_PATH:-}
+  fi
+fi
+if [ -r /etc/issue ]; then
+  if grep 'SUSE Linux Enterprise Server 11 SP1' /etc/issue > /dev/null 2>&1; then
+    LD_LIBRARY_PATH=/usr/local/lib64/:/usr/local/lib/:${LD_LIBRARY_PATH:-}
+  fi
+fi
+
 export CC CXX LD_LIBRARY_PATH
 
 CFLAGS=${CFLAGS:-"-O2"}
@@ -116,9 +135,9 @@ last_stage="galera"
 gainroot=""
 TARGET=${TARGET:-""} # default target
 
-while test $# -gt 0
+while test $# -gt 0 
 do
-    case $1 in
+    case $1 in 
         --stage)
             initial_stage=$2
             shift
@@ -179,17 +198,8 @@ do
         --sb)
             SKIP_BUILD="yes"
             ;;
-        --cmake)
-            CMAKE="yes"
-            SCONS="no"
-            ;;
-        --co)
-            CMAKE_OPTS="$CMAKE_OPTS $2"
-            shift
-            ;;
         --scons)
             SCONS="yes"
-            CMAKE="no"
             ;;
         --so)
             SCONS_OPTS="$SCONS_OPTS $2"
@@ -203,6 +213,9 @@ do
             DEBUG_LEVEL=$2
             shift
             ;;
+        -a|--archive)
+            ARCHIVE="yes"   # Generate source archive
+            ;;
         *)
             if test ! -z "$1"; then
                echo "Unrecognized option: $1"
@@ -214,32 +227,6 @@ do
     shift
 done
 
-# check whether sudo accepts -E to preserve environment
-if [ "$PACKAGE" == "yes" ]
-then
-    if which dpkg >/dev/null 2>&1
-    then
-        echo "Error: Package build not supported on Debian, use dpkg-buildpackage"
-        exit 1
-    fi
-    echo "testing sudo"
-    if sudo -E $true >/dev/null 2>&1
-    then
-        echo "sudo accepts -E"
-        SUDO="sudo -E"
-    else
-        echo "sudo does not accept param -E"
-        if [ $(id -ur) != 0 ]
-        then
-            echo "error, must build as root"
-            exit 1
-        else
-            echo "I'm root, can continue"
-            SUDO=""
-        fi
-    fi
-fi
-
 if [ "$OPT"   == "yes" ]; then CONFIGURE="yes";
    conf_flags="--disable-debug --disable-dbug";
 fi
@@ -248,12 +235,22 @@ if [ -n "$WITH_SPREAD" ]; then CONFIGURE="yes"; fi
 
 if [ "$CONFIGURE" == "yes" ] && [ "$SCONS" != "yes" ]; then SCRATCH="yes"; fi
 
+#A workaround to fix MDEV-8249 on Centos 5.
+if [ -r /etc/redhat-release ]; then
+  if grep 'CentOS release 5' /etc/redhat-release > /dev/null 2>&1; then
+    DEBUG="yes"
+  fi
+fi
+
 # Be quite verbose
 #set -x
 
 # Build process base directory
 build_base=${GALERA_SRC:-$(cd $(dirname $0)/..; pwd -P)}
 
+# Use $VERSION if one is not specified via --release
+RELEASE=${RELEASE:-$VERSION}
+
 get_arch()
 {
     if ! [ -z "$TARGET" ]
@@ -284,11 +281,12 @@ get_arch()
 build_packages()
 {
     local PKG_DIR=$build_base/scripts/packages
-    pushd $PKG_DIR
 
     local ARCH=$(get_arch)
     local WHOAMI=$(whoami)
 
+    local RET=0
+
     export BUILD_BASE=$build_base
     export GALERA_VER=$RELEASE
 
@@ -302,23 +300,39 @@ build_packages()
     local STRIP_OPT=""
     [ "$NO_STRIP" == "yes" ] && STRIP_OPT="-g"
 
-    $SUDO rm -rf $ARCH
+    rm -rf $ARCH
 
     set +e
-    if [ "$OS" == "FreeBSD" ]; then
-        if test "$NO_STRIP" != "yes"; then
-            strip $build_base/{garb/garbd,libgalera_smm.so}
+    if [ $DEBIAN -ne 0 ]; then # build DEB
+        debian_version="$(lsb_release -sc)"
+
+        # Adjust compat for older platforms
+        test "$debian_version" != "lucid" || echo 7 > debian/compat
+        test "$debian_version" != "squeeze" || echo 8 > debian/compat
+
+        dch -m -D "$debian_version" --force-distribution -v "$GALERA_VER-$debian_version" "Version upgrade"
+        # -d : Do not check build dependencies and conflicts.
+        DEB_BUILD_OPTIONS="version=$GALERA_VER revno=$GALERA_REV parallel=$JOBS nostrip" dpkg-buildpackage -us -uc -b -d
+        RET=$?
+    else
+        pushd $PKG_DIR
+        if [ "$OS" == "FreeBSD" ]; then
+            if test "$NO_STRIP" != "yes"; then
+                strip $build_base/{garb/garbd,libgalera_smm.so}
+            fi
+            ./freebsd.sh $GALERA_VER
+        else # build RPM
+            ./rpm.sh $GALERA_VER
         fi
-        ./freebsd.sh $GALERA_VER
-    else # build RPM
-        ./rpm.sh $GALERA_VER
+        RET=$?
+        popd
     fi
-    local RET=$?
 
     set -e
 
-    popd
-    if [ "$OS" == "FreeBSD" ]; then
+    if [ $DEBIAN -ne 0 ]; then
+        echo Debian - Do nothing
+    elif [ "$OS" == "FreeBSD" ]; then
         mv -f $PKG_DIR/*.tbz ./
     else
         mv -f $PKG_DIR/*.rpm ./
@@ -326,6 +340,14 @@ build_packages()
     return $RET
 }
 
+gen_source_archive()
+{
+  SOURCE_DIR="galera-$VERSION"
+  rm -rf "$SOURCE_DIR.tar.gz" "$SOURCE_DIR"
+  git checkout-index -a --prefix="$SOURCE_DIR/"
+  tar czf "$SOURCE_DIR.tar.gz" "$SOURCE_DIR"
+}
+
 build_source()
 {
     local module="$1"
@@ -364,56 +386,22 @@ build_sources()
     echo $PWD/$ret
 }
 
+GALERA_VER=$RELEASE
+
 pushd "$build_base"
-#GALERA_REV="$(svnversion | sed s/\:/,/g)"
-#if [ "$GALERA_REV" == "exported" ]
-#then
-    GALERA_REV=$(git log --pretty=oneline | wc -l) || \
-    GALERA_REV=$(bzr revno --tree -q)              || \
-    GALERA_REV=$(svn info >&/dev/null && svnversion | sed s/\:/,/g) || \
-    GALERA_REV="XXXX"
-    # trim spaces (sed is not working on Solaris, so using bash built-in)
-    GALERA_REV=${GALERA_REV//[[:space:]]/}
-#fi
+GALERA_REV=$(git log --pretty=oneline | wc -l) || \
+GALERA_REV=$(bzr revno --tree -q)              || \
+GALERA_REV=$(svn info >&/dev/null && svnversion | sed s/\:/,/g) || \
+GALERA_REV="a3edd"
+# trim spaces (sed is not working on Solaris, so using bash built-in)
+GALERA_REV=${GALERA_REV//[[:space:]]/}
+popd
 
-if [ -z "$RELEASE" ]
-then
-    source GALERA_VERSION
-    RELEASE="$GALERA_VERSION_WSREP_API.$GALERA_VERSION_MAJOR.$GALERA_VERSION_MINOR"
-fi
-
-if [ "$CMAKE" == "yes" ] # Build using CMake
-then
-    cmake_args="$CMAKE_OPTS -DGALERA_REVISION=$GALERA_REV"
-    [ -n "$TARGET"        ] && \
-        echo "WARN: TARGET=$TARGET ignored by CMake build"
-    [ -n "$RELEASE"       ] && \
-        echo "WARN: RELEASE=$RELEASE ignored by CMake build"
-    [ "$DEBUG" == "yes"   ] && cmake_args="$cmake_args -DCMAKE_BUILD_TYPE=Debug"
-    [ -n "$EXTRA_SYSROOT" ] && \
-        echo "EXTRA_SYSROOT=$EXTRA_SYSROOT ignored by CMake build"
-
-    if [ "$SCRATCH" == "yes" ]
-    then
-        (cd $build_base && make clean) || :
-        rm -f $build_base/CMakeCache.txt
-        cmake $cmake_args $build_base
-    fi
-
-    if [ "$SKIP_BUILD" != "yes" ]
-    then
-        make -j $JOBS VERBOSE=1
-    fi
-
-    if [ $RUN_TESTS ]
-    then
-        make test ARGS=-V
-    fi
-elif [ "$SCONS" == "yes" ] # Build using Scons
+if [ "$SCONS" == "yes" ] # Build using Scons
 then
     # Scons variant dir, defaults to GALERA_SRC
     export SCONS_VD=$build_base
-    scons_args="-C $build_base revno=$GALERA_REV tests=$RUN_TESTS strict_build_flags=1"
+    scons_args="-C $build_base version=$GALERA_VER revno=$GALERA_REV tests=$RUN_TESTS strict_build_flags=1"
 
     [ -n "$TARGET"        ] && scons_args="$scons_args arch=$TARGET"
     [ -n "$RELEASE"       ] && scons_args="$scons_args version=$RELEASE"
@@ -435,6 +423,12 @@ elif test "$SKIP_BUILD" == "no"; then # Build using autotools
     exit 1
 fi # SKIP_BUILD / SCONS
 
+if test "$ARCHIVE" == "yes"
+then
+    echo "Generating source archive ..."
+    gen_source_archive
+fi
+
 if test "$PACKAGE" == "yes"
 then
     build_packages
@@ -445,4 +439,4 @@ then
     build_sources
 fi
 
-popd # $build_base
+
